# -*- coding: utf-8 -*-
import logging
from typing import Dict

from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured

from bkpaas_auth.conf import bkauth_settings as conf
from bkpaas_auth.core.exceptions import ServiceError
from bkpaas_auth.core.http import http_get
from bkpaas_auth.core.user_info import BkUserInfo, RtxUserInfo

logger = logging.getLogger(__name__)


def _get_app_credentials() -> Dict[str, str]:
    """Get app credentials to verify app, which is required for requesting user info API"""
    if conf.TOKEN_APP_CODE and conf.TOKEN_SECRET_KEY:
        return {'bk_app_code': conf.TOKEN_APP_CODE, 'bk_app_secret': conf.TOKEN_SECRET_KEY}
    raise ImproperlyConfigured("BKAUTH_TOKEN_APP_CODE and BKAUTH_TOKEN_SECRET_KEY not set")


def _get_and_cache_user_info(cache_key, user_params, response_ok_checker):
    """Get user info from cache, or fetch from api and cache

    :param dict user_params: username param key to username map, it may be different in different systems
    :param callable response_ok_checker: determine get user is successful
    """
    try:
        cached_result = cache.get(cache_key)
    except Exception as e:
        # Cache is not usable due to some reason including different pickle protocols
        # between different Python versions.
        logger.warning(f"unable to get user info from cache: {e}")
        cached_result = None

    if cached_result:
        return cached_result

    params = dict(user_params, **_get_app_credentials())
    is_success, result = http_get(conf.TOKEN_USER_INFO_ENDPOINT, params=params)
    if not is_success:
        raise ServiceError('Unable to get user info')

    if not response_ok_checker(result):
        logger.error(
            f'Get user info fail, url: {conf.TOKEN_USER_INFO_ENDPOINT}, params: {params}, response: {result}',
        )
        return

    # 获取用户信息成后才缓存数据
    cache.set(cache_key, result, timeout=86400)

    return result


def get_rtx_user_info(username):
    """Get RTX user info by given RTX username. For better performance, this function
    will try to cache the result for 86400 seconds(1 day).

    :param str username: RTX username
    """

    def response_ok_checker(result):
        return result['result']

    cache_key = f'bkauth::rests::get_rtx_user_info::{username}'
    result = _get_and_cache_user_info(cache_key, {'login_name': username}, response_ok_checker)

    if result and response_ok_checker(result):
        return RtxUserInfo(**result['data'])

    return None


def get_bk_user_info(username):
    """Get BK user info by given BK username. For better performance, this function
    will try to cache the result for 86400 seconds(1 day).

    :param str username: BK username
    """

    def response_ok_checker(result):
        return result['code'] == 0

    cache_key = f'bkauth::rests::get_bk_user_info::{username}'
    result = _get_and_cache_user_info(cache_key, {'bk_username': username}, response_ok_checker)

    if result and response_ok_checker(result):
        return BkUserInfo(**result['data'])

    return None



# -*- coding: utf-8 -*-
import logging
import pickle
import time
from typing import Dict

from django.conf import settings
from django.contrib import auth
from django.http import HttpRequest
from django.utils.deprecation import MiddlewareMixin
from django.utils.encoding import force_text

from bkpaas_auth.backends import UniversalAuthBackend

logger = logging.getLogger(__name__)


class CookieLoginMiddleware(MiddlewareMixin):
    """Call auth.login when user credential cookies changes"""

    def process_request(self, request):
        assert hasattr(request, 'session'), (
            "The CookieLoginMiddleware requires session middleware "
            "to be installed. Edit your MIDDLEWARE%s setting to insert "
            "'django.contrib.sessions.middleware.SessionMiddleware' before "
            "'bkpaas_auth.middlewares.CookieLoginMiddleware'."
        ) % ("_CLASSES" if settings.MIDDLEWARE is None else "")

        backend = UniversalAuthBackend()
        credentials = backend.get_credentials(request)

        # No credentials, call logout
        if not credentials:
            auth.logout(request)
            return self.get_response(request)

        if self.should_authenticate(request, backend, credentials):
            self.authenticate_and_login(request, credentials)

        return self.get_response(request)

    def should_authenticate(
        self, request: HttpRequest, backend: UniversalAuthBackend, credentials: Dict[str, str]
    ) -> bool:
        """Decide whether to re-authenticate current credentials or not"""
        # Force re-login if credentials is different from last time
        credentials_been_modified = credentials != request.session.get('auth_credentials', {})
        if credentials_been_modified:
            return True

        # Force re-login if token is empty or obsolete
        token = backend.get_token_from_session(request)
        return token is None

    def authenticate_and_login(self, request: HttpRequest, credentials: Dict[str, str]):
        """Authenticate given credentials and do login(or logout if credentials is invalid)

        :params request: Current request object
        :params credentials: user credentials, such as uin/skey pair
        """
        logger.debug('Authenticating credentials...')
        user = auth.authenticate(request=request, auth_credentials=credentials)
        if user is None or not user.is_authenticated:
            logger.info('Authentication failed, logout.')
            auth.logout(request)
            return

        backend = auth.load_backend(user.backend)
        if not isinstance(backend, UniversalAuthBackend):
            logger.info("User is not validate by UniversalAuthBackend, skip login processes.")
            return

        logger.debug('Authentication finished, username: %s', user.username)
        request.session['provider_type'] = user.provider_type.value
        request.session['bkpaas_user_id'] = user.bkpaas_user_id
        request.session['bkpaas_authenticated_at'] = time.time()
        request.session['auth_credentials'] = credentials
        # python3 compatibility
        request.session['user_token'] = force_text(pickle.dumps(user.token), 'latin1')

        # Calling `auth.login` will rotate CSRF token and modify user session, only do this when the authenticated
        # user was different with the user stored in session. Otherwise CSRF token validation may fail due to the
        # rotation.
        if getattr(request, "user", None) != user:
            auth.login(request, user)



# -*- coding: utf-8 -*-
"""Exceptions for bkpaas_auth module
"""


class ServiceError(Exception):
    """Login or Token service is not available"""


class InvalidSkeyError(Exception):
    """Invalid uin/skey given"""


class InvalidTokenCredentialsError(Exception):
    """When invalid credentials are given when exchange access token"""



# -*- coding: utf-8 -*-
from dataclasses import dataclass, field, fields

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.test.signals import setting_changed


def get_settings(name, prefix='BKAUTH_', raise_if_missing=False, default=None):
    """Get a settings value, first in django settings, then find in ConfFixture

    :param str prefix: prefix string using by key in django settings
    :param bool raise_if_missing: raise exception if settings was not set
    """

    def factory():
        name_in_settings = prefix + name
        value = getattr(settings, name_in_settings, default)
        if raise_if_missing and value is None:
            raise ImproperlyConfigured('You must set {} in order to use bkpaas_auth'.format(name_in_settings))
        return value

    return factory


@dataclass
class Settings:

    # 用户登录态认证类型，如 bk_token
    BACKEND_TYPE: str = field(default_factory=get_settings('BACKEND_TYPE'))
    # 验证用户登录态的 API，如 蓝鲸统一登录校验登录态的 API
    USER_COOKIE_VERIFY_URL: str = field(default_factory=get_settings('USER_COOKIE_VERIFY_URL'))

    # 获取用户详情的 API，如中文名、邮箱等，且必须提供应用鉴权信息
    TOKEN_USER_INFO_ENDPOINT: str = field(default_factory=get_settings('TOKEN_USER_INFO_ENDPOINT'))
    TOKEN_APP_CODE: str = field(default_factory=get_settings('TOKEN_APP_CODE'))
    TOKEN_SECRET_KEY: str = field(default_factory=get_settings('TOKEN_SECRET_KEY'))

    LOGIN_TOKEN_EXPIRE_IN: int = field(default_factory=get_settings('LOGIN_TOKEN_EXPIRE_IN', default=24 * 60 * 60))
    SESSION_TIMEOUT: int = field(default_factory=get_settings('SESSION_TIMEOUT', default=5 * 60))

    # 请求第三方 API 设置
    REQUESTS_VERIFY: bool = field(default_factory=get_settings('REQUESTS_VERIFY', default=False))
    REQUESTS_CERT: str = field(default_factory=get_settings('REQUESTS_CERT'))

    # Test data, optional
    USE_MOCKED_USER_INFO: bool = field(default_factory=get_settings('USE_MOCKED_USER_INFO', default=False))
    MOCKED_USER_NAME: str = field(default_factory=get_settings('MOCKED_USER_NAME', default=''))

    def reload(self):
        for f in fields(self):
            setattr(self, f.name, f.default_factory())


bkauth_settings = Settings()


def reload_settings(*args, **kwargs):
    setting: str = kwargs['setting']
    if setting.startswith("BKAUTH_"):
        bkauth_settings.reload()


setting_changed.connect(reload_settings)



# -*- coding: utf-8 -*-
from enum import IntEnum


class ProviderType(IntEnum):
    UIN = 1
    RTX = 2
    BK = 3
    DATABASE = 9

    def get_id_prefix(self):
        return '{0:02d}'.format(self.value)



# -*- coding: utf-8 -*-
__version__ = "2.0.3"


def get_user_by_user_id(user_id: str, username_only: bool = True):
    """Get a user object from given user_id"""
    from bkpaas_auth.core.constants import ProviderType
    from bkpaas_auth.core.encoder import user_id_encoder
    from bkpaas_auth.core.services import get_bk_user_info, get_rtx_user_info
    from bkpaas_auth.models import User

    provider_type, username = user_id_encoder.decode(user_id)
    user = User(token=None, provider_type=ProviderType(provider_type), username=username)
    if username_only:
        return user

    # Request third party service to get info other than username
    if provider_type == ProviderType.RTX:
        user_info = get_rtx_user_info(username)
        user_info.provide(user)
        return user
    elif provider_type == ProviderType.BK:
        user_info = get_bk_user_info(username)
        user_info.provide(user)
        return user
    else:
        raise ValueError('ProviderType is not supported yet!')



# -*- coding: utf-8 -*-
"""
We need the request object to get the user, so we'll slightly modify the
existing django.contrib.auth.get_user method. To do so we update the
auth middleware to point to our overridden method.

Calling the "patch_middleware_get_user" method somewhere like our urls.py
file takes care of hooking it in appropriately.
"""
import logging

from django.contrib import auth
from django.contrib.auth import middleware, models

logger = logging.getLogger(__name__)


def middleware_get_user(request):
    if not hasattr(request, "_cached_user"):
        request._cached_user = get_user(request)
    return request._cached_user


def get_user(request):
    try:
        user_id = request.session[auth.SESSION_KEY]
        backend_path = request.session[auth.BACKEND_SESSION_KEY]
        backend = auth.load_backend(backend_path)
        backend.request = request
        user = backend.get_user(user_id) or models.AnonymousUser()
    except KeyError:
        user = models.AnonymousUser()
    except ImportError as e:
        logger.exception(f"get an anonymous user, error:{e}")
        user = models.AnonymousUser()
    return user


def patch_middleware_get_user():
    middleware.get_user = middleware_get_user
    auth.get_user = get_user



# -*- coding: utf-8 -*-
from typing import Dict

from django.contrib.auth import models
from django.db import models as db_models

from bkpaas_auth.core.constants import ProviderType
from bkpaas_auth.core.encoder import user_id_encoder


class AbstractUserWithProvider(models.AbstractBaseUser, models.AnonymousUser):
    """Basic user with provider type"""

    bkpaas_user_id = db_models.CharField(primary_key=True, max_length=255)
    USERNAME_FIELD = 'bkpaas_user_id'
    USERINFO_FIELDS = ('nickname', 'chinese_name', 'avatar_url', 'email', 'phone')

    def __init__(self, provider_type, username):
        if not provider_type:
            self.bkpaas_user_id = '-1'
        elif provider_type not in ProviderType:
            raise ValueError('Invalid provider_type given!')
        else:
            self.bkpaas_user_id = user_id_encoder.encode(provider_type, username)

        self.provider_type = provider_type
        self.username = username
        self.password = None
        # Set user info fields to default value: None
        self.update_user_info({}, overwrite_all=True)

    def update_user_info(self, info_dict: Dict, overwrite_all=False):
        """Update current user info by dict

        :param overwrite_all: if True, will set emitted field to None if that field is not
            provided by info_dict
        """
        for field in self.USERINFO_FIELDS:
            try:
                value = info_dict[field]
                setattr(self, field, value)
            except KeyError:
                if overwrite_all:
                    setattr(self, field, None)

    def save(*args, **kwargs):
        pass

    def delete(*args, **kwargs):
        pass

    @property
    def is_authenticated(self):
        return True

    @property
    def is_anonymous(self):
        return not (self.is_authenticated)

    def __eq__(self, other):
        if isinstance(other, models.AnonymousUser):
            return self.is_anonymous
        return super(AbstractUserWithProvider, self).__eq__(other)

    class Meta:
        abstract = True
        app_label = 'bkpaas_auth'


class BasicUser(AbstractUserWithProvider):
    """A basic user object with no extra stuffs"""


class User(AbstractUserWithProvider):
    """Blueking User Model provided by external user systems: Wechat Work or Uin"""

    def __init__(self, token=None, provider_type=None, username=None, **info_fields):
        super().__init__(provider_type, username)

        self.update_user_info(info_fields, overwrite_all=True)
        # Use chinesename as nickname
        if not self.nickname and self.chinese_name:
            self.nickname = self.chinese_name
        self.token = token

    @property
    def is_authenticated(self):
        # If self.token has expired, user is considered to be expired too
        # This will force user to re-login again.
        return bool(self.token and not self.token.expired())

    @property
    def is_anonymous(self):
        return not self.is_authenticated


class DatabaseUser(AbstractUserWithProvider):
    """Blueking User Model provided by external database"""

    provider_type = ProviderType.DATABASE

    @classmethod
    def from_db_obj(cls, user):
        obj = cls(cls.provider_type, username=user.username)
        obj._db_object = user
        return obj

    class Meta:
        app_label = 'bkpaas_auth'



# -*- coding: utf-8 -*-
"""
http基础方法

Rules:
1. POST/DELETE/PUT: json in
2. GET带参数，HEAD不带参数
3. 所有请求 json out，如果resp.json报错, 则是接口问题
"""
import json
import logging
from typing import Tuple, Union

import requests

from bkpaas_auth.conf import bkauth_settings

logger = logging.getLogger(__name__)

request_adapter = requests.adapters.HTTPAdapter(pool_connections=20, pool_maxsize=20)


def get_requests_session():
    """Return an empty requests session, use the function to reuse HTTP connections"""
    session = requests.session()
    session.mount("http://", request_adapter)
    session.mount("https://", request_adapter)
    session.verify = bkauth_settings.REQUESTS_VERIFY
    session.cert = bkauth_settings.REQUESTS_CERT
    return session


def _http_request(method: str, url: str, **kwargs) -> Tuple[bool, Union[None, dict, list]]:
    session = get_requests_session()
    try:
        resp = session.request(method, url, **kwargs)
    except requests.exceptions.RequestException:
        logger.exception("http request error! method: %s, url: %s, kwargs: %s", method, url, kwargs)
        return False, None

    logger.debug("request method: %s, url: %s, kwargs: %s", method, url, kwargs)

    try:
        return True, resp.json()
    except json.decoder.JSONDecodeError:
        logger.exception(
            "response json error! method: %s, url: %s, kwargs: %s, response.status_code: %s, response.text: %s",
            method,
            url,
            kwargs,
            resp.status_code,
            resp.text,
        )
        return False, None


def http_get(url: str, **kwargs) -> Tuple[bool, Union[None, dict, list]]:
    return _http_request(method="GET", url=url, **kwargs)


def http_post(url: str, **kwargs) -> Tuple[bool, Union[None, dict, list]]:
    return _http_request(method="POST", url=url, **kwargs)


def http_put(url: str, **kwargs) -> Tuple[bool, Union[None, dict, list]]:
    return _http_request(method="PUT", url=url, **kwargs)


def http_delete(url: str, **kwargs) -> Tuple[bool, Union[None, dict, list]]:
    return _http_request(method="DELETE", url=url, **kwargs)



# -*- coding: utf-8 -*-
"""Access token for blueking
"""
import datetime
import logging

from django.utils.timezone import now

from bkpaas_auth.conf import bkauth_settings
from bkpaas_auth.core.constants import ProviderType
from bkpaas_auth.core.exceptions import InvalidTokenCredentialsError, ServiceError
from bkpaas_auth.core.http import http_get
from bkpaas_auth.core.user_info import BkUserInfo, RtxUserInfo, UserInfo
from bkpaas_auth.models import User

logger = logging.getLogger(__name__)


class AbstractRequestBackend:
    def request_username(self, **credentials):
        """Get username through credentials"""


class TokenRequestBackend(AbstractRequestBackend):

    provider_type = ProviderType.BK

    def request_username(self, **credentials):
        """Get username through credentials"""
        is_success, resp = http_get(bkauth_settings.USER_COOKIE_VERIFY_URL, params=credentials, timeout=10)
        if not is_success:
            raise ServiceError('unable to fetch token services')

        # API 返回格式为：{"result": true, "code": 0, "message": "", "data": {"bk_username": "xxx"}}
        if resp.get('code') != 0:
            logger.debug(
                f'Get user fail, url: {bkauth_settings.USER_COOKIE_VERIFY_URL}, '
                f'params: {credentials}, response: {resp}'
            )
            raise InvalidTokenCredentialsError('Invalid credentials given')
        return resp["data"]["bk_username"]


class RequestBackend(AbstractRequestBackend):

    provider_type = ProviderType.RTX

    def request_username(self, **credentials):
        """Get username through credentials"""
        is_success, resp = http_get(bkauth_settings.USER_COOKIE_VERIFY_URL, params=credentials, timeout=10)
        if not is_success:
            raise ServiceError('unable to fetch token services')

        # API 返回格式为：{"msg": "", "data": {"username": "xxx"}, "ret": 0}
        if resp.get('ret') != 0:
            logger.debug(
                f'Get user fail, url: {bkauth_settings.USER_COOKIE_VERIFY_URL}, '
                f'params: {credentials}, response: {resp}'
            )
            raise InvalidTokenCredentialsError('Invalid credentials given')
        return resp["data"]["username"]


class LoginToken:
    """Access token object"""

    token_timeout_margin = 300

    def __init__(self, login_token=None, expires_in=None):
        assert login_token, 'Must provide token string'
        assert expires_in, 'Must provide expires_in seconds'
        self.login_token = login_token
        self.expires_at = now() + datetime.timedelta(seconds=expires_in)
        self.issued_at = now()
        self.user_info = UserInfo(username='AnonymousUser')

    def __str__(self):
        return 'token: {} expires_at: {}'.format(self.login_token, self.expires_at)

    def expired(self):
        return self.expires_at < now()

    def make_user(self, provider_type):
        self.user_info.provider_type = provider_type
        return create_user_from_token(self)


def mocked_create_user_from_token(
    token: LoginToken, provider_type: int = ProviderType.RTX, username: str = bkauth_settings.MOCKED_USER_NAME
) -> User:
    """Mocked create_user function, only for temporary use"""
    if provider_type == ProviderType.RTX:
        token.user_info = RtxUserInfo(
            LoginName=username,
            ChineseName=username,
        )
    elif provider_type == ProviderType.BK:
        token.user_info = BkUserInfo(bk_username=username, chname=username, email='', phone='')
    else:
        raise ValueError('Invalid provider_type given.')
    return create_user_from_token(token)


def create_user_from_token(token: LoginToken) -> User:
    """Create a user object from user info dict"""
    user = User(token=token)
    return token.user_info.provide(user)



# -*- coding: utf-8 -*-
import random
import string

DFT_RANDOM_CHARACTER_SET = string.ascii_lowercase + string.digits


def generate_random_token(length=30, chars=DFT_RANDOM_CHARACTER_SET):
    """Generates a non-guessable OAuth token

    OAuth (1 and 2) does not specify the format of tokens except that they
    should be strings of random characters. Tokens should not be guessable
    and entropy when generating the random characters is important. Which is
    why SystemRandom is used instead of the default random.choice method.
    """
    rand = random.SystemRandom()
    return ''.join(rand.choice(chars) for x in range(length))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bkpaas_auth', 'bkpaas_auth.core']

package_data = \
{'': ['*']}

install_requires = \
['cryptography', 'django>=1.11,<4.0', 'requests', 'six']

setup_kwargs = {
    'name': 'bkpaas-auth',
    'version': '2.0.3',
    'description': 'User authentication django app for blueking internal projects',
    'long_description': '# bkpaas-auth\n\n蓝鲸 PaaS 平台内部服务使用的用户鉴权模块。\n\n## 版本历史\n\n详见 `CHANGES.md`。\n\n## 开发指南\n\n### 发布包\n\n- 在 `bkpaas_auth/__init__.py`  文件中更新 `__version__`\n- 在 `setup.py` 文件中更新 `version`\n- 在 `pyproject.toml` 中更新 `version`\n- 在 `CHANGES.md` 中添加对应的版本日志\n- 执行 `poetry build` 命令在 dist 目录下生成当前版本的包。然后执行 `twine upload dist/* --repository-url {pypi_address} --username {your_name} --password {your_token}` 将其上传到 pypi 服务器上。\n\n### 关于 setup.py\n\n虽然在 [PEP 517](https://python-poetry.org/docs/pyproject/#poetry-and-pep-517) 规范里，Python 包不再需要 `setup.py` 文件。但真正少了 `setup.py` 文件后，会发现有些功能就没法正常使用，比如 pip 的可编辑安装模式、tox 等（[相关文档](https://github.com/python-poetry/poetry/issues/761)）。所以我们仍然需要它。\n\n为了避免维护重复的 `pyproject.toml` 和 `setup.py` 文件，我们使用了 [dephell](https://github.com/dephell/dephell) 工具来自动生成 `setup.py` 文件。\n\n- 安装 dephell\n- 在根目录执行 `dephell deps convert --from pyproject.toml --to setup.py`\n\n## 使用指南\n1. 更新 settings：\n```python\nINSTALLED_APPS = [\n    ...\n    \'bkpaas_auth\',\n    ...\n]\n\nMIDDLEWARE = [\n    ...\n    \'bkpaas_auth.middlewares.CookieLoginMiddleware\',\n    ...\n]\n\nAUTHENTICATION_BACKENDS = [\n    # [推荐] 使用内置的虚拟用户类型，不依赖于数据库表.\n    \'bkpaas_auth.backends.UniversalAuthBackend\',\n    # 如果项目需要保留使用数据库表的方式来设计用户模型, 则需要使用 DjangoAuthUserCompatibleBackend\n    # \'bkpaas_auth.backends.DjangoAuthUserCompatibleBackend\',\n]\n\n# 使用 bkpaas_auth 内置的基于内存的用户模型\n# 如果项目需要保留使用数据库表的方式来设计用户模型, 可阅读 「关于AUTH_USER_MODEL」的部分说明\nAUTH_USER_MODEL = \'bkpaas_auth.User\'\n\n# 用户登录态认证类型\nBKAUTH_BACKEND_TYPE = "bk_token" # 可选值：bk_token/bk_ticket\n# 验证用户登录态的 API，如 蓝鲸统一登录校验登录态的 API\nBKAUTH_USER_COOKIE_VERIFY_URL = "http://bk-login-web/api/v3/is_login/"\n\n# [可选]`BKAUTH_DEFAULT_PROVIDER_TYPE` 的值用于 JWT 校验时获取默认的用户认证类型。\nBKAUTH_DEFAULT_PROVIDER_TYPE = \'RTX\'  # 可选值：RTX/UIN/BK，详见 ProviderType\n```\n\n2. 在 app config 中进行 patch：\n\n配置登录模块的 apps.py\n\n```python\nfrom bkpaas_auth.backends import DjangoAuthUserCompatibleBackend\nfrom bkpaas_auth.models import User\nfrom django.apps import AppConfig\n\n\nclass MyAppConfig(AppConfig):\n    name = \'my_app\'\n\n    def ready(self):\n        from bkpaas_auth.monkey import patch_middleware_get_user\n\n        patch_middleware_get_user()\n```\n\n更新 `__init__.py`，配置 default_app_config\n```\ndefault_app_config = \'xxx.apps.MyAppConfig\'\n```\n\n3. 配置日志（可选）\n在 django settings 的 LOGGING 中，为 sdk 配置 logger，如\n\n```python\nLOGGING = {\n    "handlers": {\n        "root": {\n            ...\n        },\n    },\n    "loggers": {\n        "bkpaas_auth": {\n            "handlers": ["root"],\n            "level": "WARNING",\n            "propagate": True,\n        },\n    },\n}\n```\n\n### 关于 AUTH_USER_MODEL\n\nbkpaas-auth 内置的基于内存的不依赖于数据库表的用户模型, 如果需要复用原有的用户模型, 则需要使用 `DjangoAuthUserCompatibleBackend` 作为用户校验后端.\n\n在默认情况下, `DjangoAuthUserCompatibleBackend` 会从 bkpaas-auth 获取到当前登录的用户信息, 并会根据用户信息尝试创建一个基于数据库的用户模型.\n如果你有以下诉求, 则应当继承 `DjangoAuthUserCompatibleBackend`, 自行实现具体的业务逻辑:\n\n1. 不希望自动创建基于数据库的用户模型:\n```python\n\n\nclass YourDjangoAuthUserCompatibleBackend(DjangoAuthUserCompatibleBackend):\n    create_unknown_user = False\n```\n\n2. 用户模型有与 django `auth.User` 不兼容的字段或其他需要初始化的字段:\n```python\n\nclass YourDjangoAuthUserCompatibleBackend(DjangoAuthUserCompatibleBackend):\n    def configure_user(self, db_user, bk_user: User):\n        """\n        Configure a user after creation and return the updated user.\n        """\n        ...\n        return db_user\n```\n\n#### 和 [apigw-manager](../apigw-manager) 集成\n该 SDK 可以和 apigw-manager 集成，完成网关 JWT 的校验，在 settings 中配置：\n```python\nINSTALLED_APPS += ["apigw_manager.apigw"]\nAUTHENTICATION_BACKENDS += ["bkpaas_auth.backends.APIGatewayAuthBackend"]\nMIDDLEWARE += [\n    "apigw_manager.apigw.authentication.ApiGatewayJWTGenericMiddleware",  # JWT 认证\n    "apigw_manager.apigw.authentication.ApiGatewayJWTAppMiddleware",  # JWT 透传的应用信息\n    "apigw_manager.apigw.authentication.ApiGatewayJWTUserMiddleware",  # JWT 透传的用户信息\n]\n```\n\n设置之后，通过 JWT 透传的用户态会验证后，会写入到 `request.user` 中。注意，配置了不认证用户的网关资源透传的请求，会生成一个有对应用户名的匿名用户对象（`is_authenticated` 为 `False`）。\n',
    'author': 'blueking',
    'author_email': 'blueking@tencent.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING

from bkpaas_auth.core.constants import ProviderType
from bkpaas_auth.core.encoder import user_id_encoder

if TYPE_CHECKING:
    from bkpaas_auth.models import User  # noqa


class UserInfo:
    """Base class for Userinfo"""

    provider_type: ProviderType

    def __init__(self, username):
        self.username = username

    def provide(self, user: 'User'):
        user.provider_type = self.provider_type
        user.username = self.username
        user.bkpaas_user_id = user_id_encoder.encode(self.provider_type, self.username)

        user.update_user_info(self.__dict__)
        return user

    def __eq__(self, other):
        if not isinstance(other, UserInfo):
            return False
        return self.username == other.username


class RtxUserInfo(UserInfo):
    """User info for RTX user"""

    provider_type = ProviderType.RTX
    email_suffix = "@tencent.com"

    def __init__(self, **kwargs):
        super().__init__(kwargs["LoginName"])
        self.nickname = kwargs['ChineseName']
        self.chinese_name = kwargs['ChineseName']
        self.email = f'{self.username}{self.email_suffix}'
        # 用户 API 添加了限制，没有申请特殊权限的情况下无法获取手机信息
        self.phone = kwargs.get('MobilePhoneNumber', '')
        self.avatar_url = ''

    def __eq__(self, other):
        if not isinstance(other, RtxUserInfo):
            return False
        return (self.username, self.nickname, self.chinese_name, self.email, self.phone) == (
            other.username,
            other.nickname,
            other.chinese_name,
            other.email,
            other.phone,
        )


class BkUserInfo(UserInfo):
    """User info for Bk user"""

    provider_type = ProviderType.BK

    def __init__(self, **kwargs):
        # bk_username 用户英文ID
        super().__init__(kwargs["bk_username"])
        # chname 用户中文名
        self.nickname = kwargs['chname']
        self.chinese_name = kwargs['chname']
        self.email = kwargs['email']
        self.phone = kwargs['phone']
        self.avatar_url = ''

    def __eq__(self, other):
        if not isinstance(other, BkUserInfo):
            return False
        return (self.username, self.nickname, self.chinese_name, self.email, self.phone) == (
            other.username,
            other.nickname,
            other.chinese_name,
            other.email,
            other.phone,
        )



# -*- coding: utf-8 -*-
import binascii
from typing import Tuple, Union

from cryptography.hazmat.backends.openssl.backend import GetCipherByName, backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms
from six import ensure_binary, ensure_text

from bkpaas_auth.core.constants import ProviderType


class _ARC4(algorithms.ARC4):
    """ARC4 should support key sizes bellow 40-2048 bits,
    but algorithms.ARC4 only support the kes size to be in [40, 56, 64, 80, 128, 160, 192, 256] bytes.

    In order to support the key='jdvoqu3o4', we must overwrite the field `key_sizes` to remove the restriction.
    This is the reason why we must implement another ARC4 algorithm.
    """

    key_sizes = frozenset(range(40, 2049, 4))


backend.register_cipher_adapter(_ARC4, type(None), GetCipherByName("rc4"))


class BluekingUserIdEncoder:
    """Generator for blueking user id"""

    # It is not a real secret key, just for encoding the username
    secret_key = 'jdvoqu3o4'

    def encode(self, provider_type: Union[int, ProviderType], username: Union[str, bytes]):
        """Generate a hex string used as blueking user id

        :param provider_type: See constants.ProviderType
        :param username: User uin or user rtx username
        :returns: A hex string
        """
        id_prefix = ProviderType(provider_type).get_id_prefix()

        algorithm = _ARC4(ensure_binary(self.secret_key))
        cipher = Cipher(algorithm, mode=None)
        encryptor = cipher.encryptor()
        encoded = encryptor.update(ensure_binary(username))
        return id_prefix + ensure_text(binascii.hexlify(encoded))

    def decode(self, user_id: Union[str, bytes]) -> Tuple[int, str]:
        """Decode a given bk_user_id to the combination of (provider_type, username)

        :param str user_id: Blueking user id
        :returns: (provider_type, username)
        """
        _provider_type, username = user_id[:2], user_id[2:]
        provider_type = int(_provider_type)

        algorithm = _ARC4(ensure_binary(self.secret_key))
        cipher = Cipher(algorithm, mode=None)
        decryptor = cipher.decryptor()
        decoded = decryptor.update(binascii.unhexlify(username))
        return provider_type, ensure_text(decoded)


user_id_encoder = BluekingUserIdEncoder()



# -*- coding: utf-8 -*-
import logging

from django.utils.timezone import now

from bkpaas_auth.conf import bkauth_settings

logger = logging.getLogger(__name__)


class BasePlugin:
    def get_credentials(self, request):
        """获取登入态参数"""
        raise NotImplementedError

    def validate_login_token(self, login_token):
        """校验 login_token 合法性"""
        if not login_token:
            return False

        if login_token.expired():
            return False

        if (now() - login_token.issued_at).total_seconds() > bkauth_settings.SESSION_TIMEOUT:
            return False

        return True


class BkTicketPlugin(BasePlugin):
    """Auth backend for bk_ticket"""

    def get_credentials(self, request):
        bk_ticket = request.COOKIES.get('bk_ticket')

        if bk_ticket:
            return {
                'bk_ticket': bk_ticket,
            }

        return None


class BkTokenPlugin(BasePlugin):
    """Auth backend for bk_token"""

    def get_credentials(self, request):
        bk_token = request.COOKIES.get('bk_token')

        if bk_token:
            return {'bk_token': bk_token}

        return None



# -*- coding: utf-8 -*-
import logging
import pickle
from typing import Dict, Optional, Union

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest
from django.utils.encoding import force_bytes

from bkpaas_auth.conf import bkauth_settings
from bkpaas_auth.core.constants import ProviderType
from bkpaas_auth.core.exceptions import InvalidTokenCredentialsError, ServiceError
from bkpaas_auth.core.plugins import BkTicketPlugin, BkTokenPlugin
from bkpaas_auth.core.token import (
    LoginToken,
    RequestBackend,
    TokenRequestBackend,
    create_user_from_token,
    mocked_create_user_from_token,
)
from bkpaas_auth.core.user_info import UserInfo
from bkpaas_auth.core.utils import generate_random_token
from bkpaas_auth.models import User

logger = logging.getLogger(__name__)


class UniversalAuthBackend:
    """An universal cookie auth backend.

    This backend is to be used in conjunction with the ``CookieLoginMiddleware``
    found in the middleware module of this package.
    """

    request: HttpRequest

    def __init__(self):
        self.backend_type = bkauth_settings.BACKEND_TYPE
        if self.backend_type == 'bk_ticket':
            self.plugin = BkTicketPlugin()
            self.request_backend = RequestBackend()
        elif self.backend_type == 'bk_token':
            self.plugin = BkTokenPlugin()
            self.request_backend = TokenRequestBackend()
        else:
            raise ImproperlyConfigured("BKAUTH_BACKEND_TYPE not set")

    def authenticate(self, request: HttpRequest, auth_credentials: Dict) -> Optional[Union[User, AnonymousUser]]:
        try:
            username = self.request_backend.request_username(**auth_credentials)
            login_token = generate_random_token()
            token = LoginToken(
                login_token=login_token,
                expires_in=bkauth_settings.LOGIN_TOKEN_EXPIRE_IN,
            )
            token.user_info = UserInfo(username=username)
            logger.debug(f'New login token exchanged by credentials, token={login_token}')
        except InvalidTokenCredentialsError:
            logger.warning('authenticate error, invalid credentials given')
            return None
        except ServiceError:
            logger.warning('authenticate error, Error requesting third-party API service')
            return None

        return self.get_user_by_token(token)

    def get_user(self, user_id):
        """Get user from current session"""
        if not hasattr(self, 'request'):
            return

        # Try to get login_token from session
        token = self.get_token_from_session(self.request)
        if token:
            # Q: 为什么不调用 get_user_by_token?
            # A: 由于 get_user_by_token 需要访问远程服务, 但事实上在 authenticate 时, 用户信息已经被缓存到 token 对象中.
            #    为了减少网络开销, 所以直接调用 create_user_from_token.
            return create_user_from_token(token)

    def get_credentials(self, *args, **kwargs):
        return self.plugin.get_credentials(*args, **kwargs)

    def get_token_from_session(self, request: HttpRequest) -> Optional[LoginToken]:
        """Try getting token object from session"""
        if "user_token" not in request.session:
            return None

        try:
            user_token_pickled = force_bytes(request.session['user_token'], 'latin1')
            user_token: LoginToken = pickle.loads(user_token_pickled)
            logger.debug('LoginToken from session: %s', user_token)
        except Exception as error:
            logger.exception("pickle loads user_token failed, %s, %s", error, request.session['user_token'])
            return None

        # token 已经过期则不返回，否则会出现 403
        if self.plugin.validate_login_token(user_token):
            return user_token
        return None

    def get_user_by_token(self, token: LoginToken) -> User:
        """Return an user object for given token object by calling Remote User Backend.

        This Method will validate token and then fetch user info from remote backend.

        :param token: token.LoginToken object
        :returns: User/AnonymousUser object
        """
        if bkauth_settings.USE_MOCKED_USER_INFO:
            user = mocked_create_user_from_token(token)
        else:
            user = token.make_user(self.request_backend.provider_type)
        return user


class DjangoAuthUserCompatibleBackend(UniversalAuthBackend):
    """兼容 django auth.User 的 backend.

    By default, the ``authenticate`` method creates ``User`` objects for usernames
    that don't already exist in the database. Subclasses can disable this behavior by setting
    the ``create_unknown_user`` attribute to ``False``.

    Note: This backend work like ``django.contrib.auth.backends.RemoteUserBackend``.
    """

    # Create a User object if not already in the database?
    create_unknown_user = True

    def authenticate(self, request: HttpRequest, auth_credentials: Dict) -> Optional[Union[User, AnonymousUser]]:
        user = super().authenticate(request, auth_credentials)
        if user:
            user = self.connect_to_django_user(user)
        return user

    def get_user(self, user_id):
        user = super().get_user(user_id)
        if user:
            user = self.connect_to_django_user(user)
        return user

    def connect_to_django_user(self, user: User):
        """Connect bkpaas_auth.User to the UserModel in the database."""
        UserModel = get_user_model()
        if self.create_unknown_user:
            db_user, created = UserModel._default_manager.get_or_create(**{UserModel.USERNAME_FIELD: user.username})
            if created:
                logger.info("User named %s is created!", user.username)
                db_user = self.configure_user(db_user=db_user, bk_user=user)
        else:
            try:
                db_user = UserModel._default_manager.get_by_natural_key(user.username)
            except UserModel.DoesNotExist:
                logger.warning("User named %s not found!", user.username)
                db_user = None

        if db_user:
            # Set those attribute to make db_user compatible with CookieLoginMiddleware
            db_user.provider_type = user.provider_type
            db_user.bkpaas_user_id = user.bkpaas_user_id
            db_user.token = user.token
        return db_user

    def configure_user(self, db_user, bk_user: User):
        """
        Configure a user after creation and return the updated user.
        """
        default_admin_superusers = getattr(settings, 'DEFAULT_ADMIN_SUPERUSERS', [])
        if db_user.username in default_admin_superusers:
            db_user.is_active = True
            db_user.is_staff = True
            db_user.is_superuser = True

        db_user.email = bk_user.email
        db_user.save(update_fields=["is_active", "is_staff", "is_superuser", "email"])
        return db_user


class APIGatewayAuthBackend:
    """This backend is to be used in conjunction with the ``ApiGatewayJWTUserMiddleware``
    found in the middleware module of ``apigw_manager`` package.
    """

    def authenticate(self, request, api_name, bk_username, verified, **credential):
        if not verified:
            return self.make_anonymous_user(bk_username)

        return User(
            token=LoginToken('any_token', expires_in=86400),
            provider_type=self.get_provider_type(),
            username=bk_username,
        )

    def get_user(self, user_id):
        raise NotImplementedError(
            "ApiGatewayJWTUserMiddleware should be overwrite request.user, "
            "so that APIGatewayAuthBackend.get_user will never be called."
        )

    def get_provider_type(self):
        name = getattr(settings, 'BKAUTH_DEFAULT_PROVIDER_TYPE', "RTX")
        return getattr(ProviderType, name)

    def make_anonymous_user(self, bk_username=None):
        user = AnonymousUser()
        user.username = bk_username
        return user
